use {
    crate::{Entry, Item, Package, Plist, Specs},
    rayon::prelude::{IntoParallelRefIterator, ParallelIterator},
    std::{
        borrow::BorrowMut,
        env,
        error::Error,
        fs::{File, self},
        io::{self, Write},
        path::{Path, PathBuf},
        sync::{
            atomic::{AtomicUsize, Ordering},
            mpsc::Sender,
            Mutex,
        },
    },
    walkdir::{DirEntry, WalkDir},
    zstd::Encoder,
};

pub enum Message {
    MemberAdded(String),
    Success(String),
    Failure(String),
}

pub struct Creator {
    path: PathBuf,
    entries: Vec<Result<DirEntry, walkdir::Error>>,
    specs: Specs,
}

impl Creator {
    pub fn new(path: &Path, specs: Specs) -> Result<Self, io::Error> {
        let d = env::current_dir()?;
        env::set_current_dir(path)?;
        let path = path.to_path_buf();
        let entries = WalkDir::new(".").into_iter().collect::<Vec<_>>();
        env::set_current_dir(d)?;
        Ok(Self {
            path,
            entries,
            specs,
        })
    }

    pub fn create(self, outdir: &Path, sender: Sender<Message>) -> Result<(), Box<dyn Error>> {
        let d = env::current_dir()?;
        let plist = Mutex::new(Plist::default());
        let totalsize: AtomicUsize = 0.into();
        let fullname = format!(
            "{}-{}_{}",
            &self.specs.name, self.specs.version, self.specs.release
        );
        if !outdir.exists() {
            fs::create_dir_all(&outdir)?;
        }
        let outdir = outdir.canonicalize()?;
        let mut archive = outdir.clone();
        archive.push(&fullname);
        archive.set_extension("tar.zst");
        let fd = File::create(&archive)?;
        let writer = Mutex::new(Encoder::new(fd, 0)?);
        let sender = Mutex::new(sender);
        env::set_current_dir(&self.path)?;
        self.entries
            .par_iter()
            .filter(|x| x.is_ok())
            .map(|x| x.as_ref().unwrap())
            .filter(|x| x.path() != Path::new("."))
            .for_each(|x| {
                let sender = sender.lock().unwrap().clone();
                if let Ok(item) = Item::try_create(x.path().to_path_buf().as_path()) {
                    if let Entry::File {
                        path: _,
                        sha256sum: _,
                        mode: _,
                        size,
                    } = &item.entry
                    {
                        totalsize.fetch_add(*size, Ordering::Release);
                    }
                    let path = match item.entry.clone() {
                        Entry::File {
                            path,
                            sha256sum: _,
                            mode: _,
                            size: _,
                        } => path.clone(),
                        Entry::Link { path, target: _ } => path.clone(),
                        Entry::Directory { path, mode: _ } => path.clone(),
                    };
                    plist.lock().unwrap().borrow_mut().entries.push(item.entry);
                    match writer.lock().unwrap().borrow_mut().write_all(&item.data) {
                        Ok(_) => sender
                            .send(Message::MemberAdded(format!(
                                "{} added to archive",
                                path.display()
                            )))
                            .expect("couldn't send message"),
                        Err(e) => sender
                            .send(Message::Failure(format!("{e}")))
                            .expect("couldn't send message"),
                    }
                } else {
                    sender
                        .send(Message::Failure(format!(
                            "Could not process DirEntry for {}",
                            x.path().display()
                        )))
                        .expect("could not send message");
                }
            });
        let mut package: Package = self.specs.into();
        package.size = totalsize.into_inner();
        let plist = plist.into_inner()?;
        package.plist = plist;
        let node = package.save_ron_and_create_tar_node(&outdir)?;
        let mut writer = writer.into_inner()?;
        writer.write_all(&node.to_vec()?)?;
        let _fd = writer.finish()?;
        let sender = sender.into_inner()?;
        sender.send(Message::Success(format!("{} saved", archive.display())))?;
        env::set_current_dir(d)?;
        Ok(())
    }
}
