use crate::Item;
use std::{fs, path::PathBuf};
use tar::Owner;
use zstd::Encoder;

mod dependency;
mod specs;

use {
    crate::{Entry, Plist, Version},
    rayon::prelude::*,
    ron::ser::{to_string_pretty, PrettyConfig},
    serde::{Deserialize, Serialize},
    std::{
        env,
        error::Error,
        fs::File,
        io::{BufWriter, Write},
        path::Path,
    },
    tar::Node,
    walkdir::WalkDir,
};

pub use {dependency::Dependency, specs::Specs};

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
pub struct User {
    pub name: String,
    pub uid: Option<u32>,
}

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
pub struct Group {
    pub name: String,
    pub gid: Option<u32>,
}

#[derive(Clone, Debug, Default, Serialize, Deserialize, PartialEq)]
/// the metadata associated with a package
pub struct Package {
    /// The name of the package minus all version information
    pub name: String,
    /// The `Version` of the package
    pub version: Version,
    /// The release number for this package
    pub release: u8,
    /// a single line description of the package
    pub description: String,
    /// a more verbose description of the package
    pub long_description: String,
    /// an optional link to an
    /// [AppStream](https://www.freedesktop.org/wiki/Distributions/AppStream/)
    /// metadata file
    pub appstream_data: Option<String>,
    /// a listing of all files, directories and symlinks which are a part of
    /// this package
    pub plist: Plist,
    /// the total size of this package
    pub size: usize,
    /// all of this package's runtime dependencies
    pub dependencies: Vec<Dependency>,
    /// an optional list of users to be created upon installation
    pub users: Option<Vec<User>>,
    /// an optional list of groups to be created upon installation
    pub groups: Option<Vec<Group>>,
    /// an optional post installation shell script to be run
    pub post_install: Option<String>,
}

impl From<Specs> for Package {
    fn from(value: Specs) -> Self {
        Package {
            name: value.name,
            version: value.version,
            release: value.release,
            description: value.description,
            long_description: value.long_description,
            appstream_data: value.appstream_data,
            dependencies: value.dependencies,
            users: value.users,
            groups: value.groups,
            post_install: value.post_install,
            ..Default::default()
        }
    }
}

impl Package {
    fn as_ron(&self) -> Result<String, ron::Error> {
        let cfg = PrettyConfig::new().struct_names(true);
        to_string_pretty(self, cfg)
    }

    pub(crate) fn save_ron_and_create_tar_node(
        &self,
        outdir: &Path,
    ) -> Result<Node, Box<dyn Error>> {
        if !outdir.exists() {
            fs::create_dir_all(outdir)?;
        }
        let mut outfile = outdir.to_path_buf();
        outfile.push("package.ron");
        let fd = File::create(&outfile)?;
        let s = self.as_ron()?;
        let mut writer = BufWriter::new(&fd);
        writer.write_all(s.as_bytes())?;
        writer.flush()?;
        println!("{} written", outfile.display());
        let node = Node::read_data_to_tar(
            s.as_bytes(),
            "package.ron",
            &outfile.metadata()?,
            Some(Owner::default()),
        )?;
        Ok(node)
    }

    /// Returns the formatted full package name including version and release strings
    pub fn fullname(&self) -> String {
        format!("{}-{}_{}", self.name, self.version, self.release)
    }

    /// Tests whether this package is an update for another
    pub fn is_upgrade(&self, other: &Self) -> bool {
        self.name == other.name && self.version > other.version
    }
}

/// Creates a package archive given a path to the included files, a package `Specs`
/// file and an output directory
///
/// ## Usage
/// ```
/// use std::path::Path;
/// use hpk::Specs;
///
/// let staging = Path::new("staged");
/// let outdir = Path::new("package");
/// let specs = Specs {
///     name: "foo".to_string(),
///     version: "0.42.0".parse().unwrap(),
///     description: "an example package".to_string(),
///     ..Default::default()
/// };
/// let result = hpk::create_package(&staging, specs, &outdir);
pub fn create_package(path: &Path, specs: Specs, outdir: &Path) -> Result<(), Box<dyn Error>> {
    if !outdir.exists() {
        fs::create_dir_all(outdir)?;
    }
    let outdir = outdir.canonicalize()?;
    env::set_current_dir(path)?;
    let mut items = WalkDir::new(".")
        .into_iter()
        .collect::<Vec<_>>()
        .par_iter()
        .filter(|x| x.is_ok())
        .filter_map(|x| Item::try_create(x.as_ref().unwrap().path().to_path_buf().as_path()).ok())
        .collect::<Vec<_>>();
    let mut plist = Plist::default();
    let mut archive = vec![];
    let mut totalsize = 0;
    while let Some(item) = items.pop() {
        if let Entry::File {
            path: _,
            sha256sum: _,
            mode: _,
            size,
        } = &item.entry
        {
            totalsize += size;
        }
        plist.entries.push(item.entry);
        archive.extend(item.data);
    }
    let mut package: Package = specs.into();
    package.plist = plist;
    package.size = totalsize;
    let node = package.save_ron_and_create_tar_node(&outdir)?.to_vec()?;
    archive.extend(node);
    let name = package.fullname();
    let mut path = PathBuf::from(&name);
    path.set_extension("tar.zst");
    let mut outfile = outdir;
    outfile.push(path);
    let fd = File::create(&outfile)?;
    let mut writer = Encoder::new(fd, 0)?;
    writer.write_all(&archive)?;
    let _fd = writer.finish()?;
    println!("{} saved", outfile.display());
    Ok(())
}
