use {
    chrono::{offset::Utc, DateTime},
    serde::{Deserialize, Serialize},
    std::{error::Error, fmt, str::FromStr},
};

mod gitrev;
mod semver;

pub use {gitrev::GitRev, semver::SemVer};

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
pub enum Version {
    Number(u32),
    SemVer {
        major: u8,
        minor: u8,
        patch: u8,
    },
    Git {
        hash: String,
        datetime: DateTime<Utc>,
    },
}

impl Default for Version {
    fn default() -> Self {
        Self::SemVer {
            major: 0,
            minor: 1,
            patch: 0,
        }
    }
}

impl fmt::Display for Version {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Number(n) => write!(f, "{n}"),
            Self::SemVer {
                major,
                minor,
                patch,
            } => {
                let v = SemVer {
                    major: *major,
                    minor: *minor,
                    patch: *patch,
                };
                write!(f, "{v}")
            }
            Self::Git { hash, datetime } => {
                let v = GitRev {
                    hash: hash.clone(),
                    datetime: *datetime,
                };
                write!(f, "{v}")
            }
        }
    }
}

impl From<SemVer> for Version {
    fn from(value: SemVer) -> Self {
        Self::SemVer {
            major: value.major,
            minor: value.minor,
            patch: value.patch,
        }
    }
}

impl From<GitRev> for Version {
    fn from(value: GitRev) -> Self {
        Self::Git {
            hash: value.hash,
            datetime: value.datetime,
        }
    }
}

impl From<u32> for Version {
    fn from(value: u32) -> Self {
        Self::Number(value)
    }
}

impl PartialOrd for Version {
    #[allow(clippy::many_single_char_names)]
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        match (self, other) {
            (Self::Number(s), Self::Number(o)) => s.partial_cmp(o),
            (
                Self::SemVer {
                    major: a,
                    minor: b,
                    patch: c,
                },
                Self::SemVer {
                    major: d,
                    minor: e,
                    patch: f,
                },
            ) => (a, b, c).partial_cmp(&(d, e, f)),
            (
                Self::Git {
                    hash: a,
                    datetime: b,
                },
                Self::Git {
                    hash: c,
                    datetime: d,
                },
            ) => (a, b).partial_cmp(&(c, d)),
            _ => None,
        }
    }
}

#[derive(Debug)]
pub struct ParseVersionError;

impl fmt::Display for ParseVersionError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "error parsing version")
    }
}

impl Error for ParseVersionError {}

impl FromStr for Version {
    type Err = ParseVersionError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if let Ok(v) = s.parse::<SemVer>() {
            Ok(v.into())
        } else if let Ok(v) = s.parse::<GitRev>() {
            Ok(v.into())
        } else if let Ok(v) = s.parse::<u32>() {
            Ok(v.into())
        } else {
            Err(ParseVersionError)
        }
    }
}
