use {
    serde::{Deserialize, Serialize},
    std::{cmp::Ordering, error::Error, fmt, num::ParseIntError, str::FromStr},
};

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct SemVer {
    pub major: u8,
    pub minor: u8,
    pub patch: u8,
}

impl fmt::Display for SemVer {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}.{}.{}", self.major, self.minor, self.patch)
    }
}

impl PartialOrd for SemVer {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        match self.major.partial_cmp(&other.major) {
            Some(Ordering::Greater) => Some(Ordering::Greater),
            Some(Ordering::Less) => Some(Ordering::Less),
            None => None,
            Some(Ordering::Equal) => match self.minor.partial_cmp(&other.minor) {
                Some(Ordering::Greater) => Some(Ordering::Greater),
                Some(Ordering::Less) => Some(Ordering::Less),
                None => None,
                Some(Ordering::Equal) => self.patch.partial_cmp(&other.patch),
            },
        }
    }
}

impl Ord for SemVer {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        match self.major.cmp(&other.major) {
            Ordering::Greater => Ordering::Greater,
            Ordering::Less => Ordering::Less,
            Ordering::Equal => match self.minor.cmp(&other.minor) {
                Ordering::Greater => Ordering::Greater,
                Ordering::Less => Ordering::Less,
                Ordering::Equal => self.patch.cmp(&other.patch),
            },
        }
    }
}

#[derive(Debug, PartialEq)]
pub struct ParseSemVerError;

impl fmt::Display for ParseSemVerError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "Error parsing SemVer")
    }
}

impl Error for ParseSemVerError {}

impl From<ParseIntError> for ParseSemVerError {
    fn from(_value: ParseIntError) -> Self {
        Self
    }
}

impl FromStr for SemVer {
    type Err = ParseSemVerError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let split = s.split('.').collect::<Vec<_>>();
        match split.len() {
            3 => {
                let major = split.first().unwrap().parse::<u8>()?;
                let minor = split.get(1).unwrap().parse::<u8>()?;
                let patch = split.get(2).unwrap().parse::<u8>()?;
                Ok(Self {
                    major,
                    minor,
                    patch,
                })
            }
            _ => Err(ParseSemVerError),
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn parse_semver() {
        assert_eq!(
            "1.0.3".parse::<SemVer>(),
            Ok(SemVer {
                major: 1,
                minor: 0,
                patch: 3
            })
        );
    }
}
